from onegov.user import User


def test_user_initials():
    user = User(username='admin')
    assert user.initials == 'A'

    user = User(username='nathan.drake@example.org')
    assert user.initials == 'ND'

    user = User(username='a.b.c.d.e.f')
    assert user.initials == 'AB'


def test_user_title():
    user = User(username='admin')
    assert user.title == 'Admin'

    user = User(username='nathan.drake@example.org')
    assert user.title == 'Nathan Drake'

    user = User(username='a.b.c.d.e.f')
    assert user.title == 'A B'



import logging
log = logging.getLogger('onegov.user')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from translationstring import TranslationStringFactory
_ = TranslationStringFactory('onegov.user')  # noqa

from onegov.user.auth import Auth, is_valid_yubikey
from onegov.user.collection import UserCollection
from onegov.user.model import User

__all__ = ['Auth', 'is_valid_yubikey', 'User', 'UserCollection']



from onegov.core.crypto import random_token
from onegov.user import log
from onegov.user.model import User
from onegov.user.errors import (
    AlreadyActivatedError,
    ExistingUserError,
    InsecurePasswordError,
    InvalidActivationTokenError,
    UnknownUserError,
)
from sqlalchemy import sql


MIN_PASSWORD_LENGTH = 8


class UserCollection(object):
    """ Manages a list of users.

    Use it like this::

        from onegov.user import UserCollection
        users = UserCollection(session)

    """

    def __init__(self, session):
        self.session = session

    def query(self):
        """ Returns a query using :class:`onegov.user.model.User`. """
        return self.session.query(User)

    def add(self, username, password, role,
            data=None, second_factor=None, active=True):
        """ Add a user to the collection.

            The arguments given to this function are the attributes of the
            :class:`~onegov.user.models.User` class with the same name.
        """
        assert username and password and role

        user = User(
            username=username,
            password=password,
            role=role,
            data=data,
            second_factor=second_factor,
            active=active
        )

        self.session.add(user)
        self.session.flush()

        return user

    def exists(self, username):
        """ Returns True if the given username exists.

        This function does not actually load a user, so it is the quickest
        way to find out if a user exists. It should be used if you don't
        care about finding out anything about the user.

        """
        query = self.session.query(sql.exists().where(
            User.username == username))

        return query.scalar()

    def by_id(self, id):
        """ Returns the user by the internal id.

        Use this if you need to refer to a user by path. Usernames are not
        the correct way, since they allow for user enumeration.

        """

        return self.query().filter(User.id == id).first()

    def by_username(self, username):
        """ Returns the user by username. """
        return self.query().filter(User.username == username).first()

    def by_username_and_password(self, username, password):
        """ Returns the user by username and password.

        Note that although the password can be empty on the user, this function
        will not query for empty passwords as an added security measure.

        Apart from that everything is fair game though, as it is not the job
        of onegov.user to enforce a password policy.

        """
        user = self.by_username(username)

        if user and password and user.is_matching_password(password):
            return user
        else:
            return None

    def register(self, username, password, request, role='member'):
        """ Registers a new user.

        The so created user needs to activated with a token before it becomes
        active. Use the activation_token in the data dictionary together
        with the :meth:`activate_with_token` function.

        """

        assert username

        # we could implement a proper password policy, but a min-length of
        # of eight characters is a good start. What we don't want is someone
        # registering a user with a password of one character.
        if len(password) < MIN_PASSWORD_LENGTH:
            raise InsecurePasswordError()

        if self.by_username(username):
            raise ExistingUserError("{} already exists".format(username))

        log.info("Registration by {} ({})".format(
            request.client_addr, username))

        return self.add(
            username=username,
            password=password,
            role=role,
            data={
                'activation_token': random_token()
            },
            active=False
        )

    def activate_with_token(self, username, token):
        """ Activates the user if the given token matches the verification
        token stored in the data dictionary.

        """
        user = self.by_username(username)

        if not user:
            raise UnknownUserError("{} does not exist".format(username))

        if user.active:
            raise AlreadyActivatedError("{} already active".format(username))

        if user.data.get('activation_token', object()) != token:
            raise InvalidActivationTokenError("{} is invalid".format(token))

        del user.data['activation_token']
        user.active = True
        self.session.flush()

    def delete(self, username):
        """ Deletes the user if it exists.

        If the user does not exist, an
        :class:`onegov.user.errors.UnknownUserError` is raised.

        """
        user = self.by_username(username)

        if not user:
            raise UnknownUserError("user {} does not exist".format(username))

        self.session.delete(user)
        self.session.flush()



import pytest

from onegov.core.crypto import RANDOM_TOKEN_LENGTH
from onegov.user import UserCollection
from onegov.user.errors import (
    AlreadyActivatedError,
    ExistingUserError,
    InsecurePasswordError,
    InvalidActivationTokenError,
    UnknownUserError,
)


def test_user_add(session):

    users = UserCollection(session)
    user = users.add('dmr@example.org', 'p@ssw0rd', 'root')

    assert users.by_username(user.username).username == 'dmr@example.org'


def test_user_exists(session):

    users = UserCollection(session)

    assert not users.exists('dmr@example.org')
    users.add('dmr@example.org', 'p@ssw0rd', 'root')
    assert users.exists('dmr@example.org')


def test_user_data(session):

    users = UserCollection(session)

    user = users.add('dmr@example.org', 'p@ssw0rd', 'root', data={
        'firstname': 'Dennis',
        'lastname': 'Ritchie'
    })

    assert user.data['firstname'] == 'Dennis'
    assert user.data['lastname'] == 'Ritchie'


def test_user_delete(session):

    users = UserCollection(session)
    users.add('dmr@example.org', 'p@ssw0rd', 'root')

    assert users.exists('dmr@example.org')
    users.delete('dmr@example.org')
    assert not users.exists('dmr@example.org')

    with pytest.raises(UnknownUserError):
        users.delete('dmr@example.org')


def test_user_password(session):

    users = UserCollection(session)
    user = users.add('AzureDiamond', 'hunter2', 'irc-user')
    assert user.password.startswith('$bcrypt-sha256$2a')
    assert user.password_hash.startswith('$bcrypt-sha256$2a')

    def may_login(username, password):
        return bool(users.by_username_and_password(username, password))

    assert may_login('AzureDiamond', 'hunter2')
    assert not may_login('AzureDiamond', 'hunter3')
    assert not may_login('AzureDiamon', 'hunter2')

    user.password = 'test'
    assert user.password.startswith('$bcrypt-sha256$2a')
    assert user.password_hash.startswith('$bcrypt-sha256$2a')
    assert may_login('AzureDiamond', 'test')

    # this is an exception, we allow for empty passwords, but we will not
    # query for them by default, see usercollection class
    user.password = ''
    assert user.password.startswith('$bcrypt-sha256$2a')
    assert user.password_hash.startswith('$bcrypt-sha256$2a')
    assert not may_login('AzureDiamon', '')


def test_register_user(session):

    users = UserCollection(session)

    class MockRequest(object):
        client_addr = '127.0.0.1'

    request = MockRequest()

    with pytest.raises(AssertionError):
        users.register(None, None, request)

    with pytest.raises(InsecurePasswordError):
        users.register('user', 'short', request)

    user = users.register('user', 'p@ssw0rd', request)
    token = user.data['activation_token']

    assert len(token) == RANDOM_TOKEN_LENGTH
    assert not user.active
    assert user.role == 'member'

    with pytest.raises(UnknownUserError):
        users.activate_with_token('waldo', 'asdf')

    with pytest.raises(InvalidActivationTokenError):
        users.activate_with_token('user', 'asdf')

    users.activate_with_token('user', token)
    assert user.active
    assert 'activation_token' not in user.data

    with pytest.raises(ExistingUserError):
        user = users.register('user', 'p@ssw0rd', request)

    with pytest.raises(AlreadyActivatedError):
        users.activate_with_token('user', token)



from onegov.form import Form
from onegov.user import _
from onegov.user.collection import UserCollection, MIN_PASSWORD_LENGTH
from wtforms import StringField, PasswordField, validators


class RegistrationForm(Form):
    """ A generic registration form for onegov.user """

    username = StringField(
        label=_("E-Mail Address"),
        validators=[validators.InputRequired(), validators.Email()]
    )

    password = PasswordField(
        label=_("Password"),
        validators=[
            validators.InputRequired(),
            validators.Length(min=MIN_PASSWORD_LENGTH, message=_(
                "The password must be at least eight characters long"
            ))
        ]
    )

    confirm = PasswordField(
        label=_("Password Confirmation"),
        validators=[validators.InputRequired(), validators.EqualTo(
            'password', message=_("Passwords must match")
        )]
    )

    # To avoid the most basic spam bots we present a field which the user
    # should not fill. We'll actually hide this field from the user. Bots
    # tend to fill out all fields. If they do they won't succeed.
    #
    # This however won't deterr a targeted attack. We'll have to see if we
    # need a more sophisticated approach in the future.
    roboter_falle = StringField(
        label=_("Please leave this field empty"),
        validators=[validators.Length(max=0, message=_(
            "The field is not empty"
        ))]
    )

    def register_user(self, request, role='member'):
        """ Registers the user using the information on the form.

        See :meth:`onegov.user.collections.UserCollection.register_user` for
        more information.

        """

        return UserCollection(request.app.session()).register(
            self.username.data, self.password.data, request)



import morepath
import transaction

from more.itsdangerous import IdentityPolicy
from onegov.user import Auth, is_valid_yubikey, UserCollection
from webtest import TestApp as Client
from unittest.mock import patch
from yubico_client import Yubico


def test_auth_login(session):
    UserCollection(session).add('AzureDiamond', 'hunter2', 'irc-user')
    auth = Auth(session=session, application_id='my-app')

    assert not auth.login(username='AzureDiamond', password='hunter1')
    assert not auth.login(username='AzureDiamonb', password='hunter2')

    identity = auth.login(username='AzureDiamond', password='hunter2')
    assert identity.userid == 'AzureDiamond'
    assert identity.role == 'irc-user'
    assert identity.application_id == 'my-app'


def test_auth_login_inactive(session):
    user = UserCollection(session).add(
        'AzureDiamond', 'hunter2', 'irc-user', active=False)

    auth = Auth(session=session, application_id='my-app')

    assert not auth.login(username='AzureDiamond', password='hunter2')

    user.active = True
    transaction.commit()

    assert auth.login(username='AzureDiamond', password='hunter2')


def test_is_valid_yubikey_otp(session):

    assert not is_valid_yubikey(
        client_id='abc',
        secret_key='dGhlIHdvcmxkIGlzIGNvbnRyb2xsZWQgYnkgbGl6YXJkcyE=',
        expected_yubikey_id='ccccccbcgujx',
        yubikey='ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded'
    )

    with patch.object(Yubico, 'verify') as verify:
        verify.return_value = True

        assert is_valid_yubikey(
            client_id='abc',
            secret_key='dGhlIHdvcmxkIGlzIGNvbnRyb2xsZWQgYnkgbGl6YXJkcyE=',
            expected_yubikey_id='ccccccbcgujh',
            yubikey='ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded'
        )


def test_auth_login_yubikey(session):
    UserCollection(session).add(
        username='admin@example.org',
        password='p@ssw0rd',
        role='admin',
        second_factor={
            'type': 'yubikey',
            'data': 'ccccccbcgujh'
        }
    )

    auth = Auth(
        session=session,
        application_id='my-app',
        yubikey_client_id='abc',
        yubikey_secret_key='dGhlIHdvcmxkIGlzIGNvbnRyb2xsZWQgYnkgbGl6YXJkcyE='
    )

    assert not auth.login(username='admin@example.org', password='p@ssw0rd')
    assert not auth.login(
        username='admin@example.org',
        password='p@ssw0rd',
        second_factor='xxxxxxbcgujhingjrdejhgfnuetrgigvejhhgbkugded'
    )

    with patch.object(Yubico, 'verify') as verify:
        verify.return_value = False

        assert not auth.login(
            username='admin@example.org',
            password='p@ssw0rd',
            second_factor='ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded'
        )

    with patch.object(Yubico, 'verify') as verify:
        verify.return_value = True

        identity = auth.login(
            username='admin@example.org',
            password='p@ssw0rd',
            second_factor='ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded'
        )

    assert identity.userid == 'admin@example.org'
    assert identity.role == 'admin'
    assert identity.application_id == 'my-app'


def test_auth_login_unnecessary_yubikey(session):
    UserCollection(session).add(
        username='admin@example.org',
        password='p@ssw0rd',
        role='admin'
    )

    auth = Auth(
        session=session,
        application_id='my-app',
    )

    # simply ignore the second factor
    assert auth.login(
        username='admin@example.org',
        password='p@ssw0rd',
        second_factor='ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded'
    )


def test_auth_logging(session, capturelog):
    UserCollection(session).add('AzureDiamond', 'hunter2', 'irc-user')
    auth = Auth(session=session, application_id='my-app')

    auth.login(username='AzureDiamond', password='hunter1')
    assert capturelog.records()[0].message \
        == "Failed login by unknown (AzureDiamond)"

    auth.login(username='AzureDiamond', password='hunter1', client='127.0.0.1')
    assert capturelog.records()[1].message \
        == "Failed login by 127.0.0.1 (AzureDiamond)"

    auth.login(username='AzureDiamond', password='hunter2', client='127.0.0.1')
    assert capturelog.records()[2].message \
        == "Successful login by 127.0.0.1 (AzureDiamond)"


def test_auth_integration(session):

    class App(morepath.App):
        pass

    @App.identity_policy()
    def get_identity_policy():
        return IdentityPolicy()

    @App.path(path='/auth', model=Auth)
    def get_auth():
        return Auth(session, application_id='my_app', to='https://abc.xyz/go')

    @App.view(model=Auth)
    def view_auth(self, request):
        return self.login_to(
            request.params.get('username'),
            request.params.get('password'),
            request
        ) or 'Error'

    @App.view(model=Auth, name='logout')
    def view_logout(self, request):
        return self.logout_to(request)

    App.commit()

    UserCollection(session).add('AzureDiamond', 'hunter2', 'irc-user')
    transaction.commit()

    client = Client(App())

    response = client.get('/auth?username=AzureDiamond&password=hunter1')
    assert response.text == 'Error'

    response = client.get('/auth?username=AzureDiamond&password=hunter2')
    assert response.status_code == 302
    assert response.location == 'http://localhost/go'
    assert response.headers['Set-Cookie'].startswith('userid=AzureDiamond')

    response = client.get('/auth/logout')
    assert response.status_code == 302
    assert response.location == 'http://localhost/go'
    assert response.headers['Set-Cookie'].startswith('userid=;')



class OnegovUserError(Exception):
    """ Base class for all errors raised by onegov.user. """
    def __init__(self, message):
        self.message = message


class UnknownUserError(OnegovUserError):
    """ Raised when a user was not found. """


class InvalidActivationTokenError(OnegovUserError):
    """ Raised when the given activation token doesn't match. """


class ExistingUserError(OnegovUserError):
    """ Raised when a new user already exists. """


class AlreadyActivatedError(OnegovUserError):
    """ Raised when a user was already activated. """


class InsecurePasswordError(OnegovUserError):
    """ Raised when a user's password is not secure enough. """

    def __init__(self):
        pass



import morepath
from onegov.core.utils import relative_url
from onegov.user import log
from onegov.user.collection import UserCollection
from yubico_client import Yubico
from yubico_client.yubico_exceptions import (
    StatusCodeError,
    SignatureVerificationError
)


def is_valid_yubikey(client_id, secret_key, expected_yubikey_id, yubikey):
    """ Asks the yubico validation servers if the given yubikey OTP is valid.

    :client_id:
        The yubico API client id.

    :secret_key:
        The yubico API secret key.

    :expected_yubikey_id:
        The expected yubikey id. The yubikey id is defined as the first twelve
        characters of any yubikey value. Each user should have a yubikey
        associated with it's account. If the yubikey value comes from a
        different key, the key is invalid.

    :yubikey:
        The actual yubikey value that should be verified.

    :return: True if yubico confirmed the validity of the key.

    """
    assert client_id and secret_key and expected_yubikey_id and yubikey
    assert len(expected_yubikey_id) == 12

    # if the yubikey doesn't start with the expected yubikey id we do not
    # need to make a roundtrip to the validation server
    if not yubikey.startswith(expected_yubikey_id):
        return False

    try:
        return Yubico(client_id, secret_key).verify(yubikey)
    except StatusCodeError as e:
        if e.status_code != 'REPLAYED_OTP':
            raise e

        return False
    except SignatureVerificationError as e:
        return False


class Auth(object):
    """ Defines a model for authentication methods like login/logout.
    Applications should use this model to implement authentication views.

    """

    identity_class = morepath.Identity

    def __init__(self, session, application_id, to='/',
                 yubikey_client_id=None, yubikey_secret_key=None):
        assert application_id  # may not be empty!

        self.session = session
        self.application_id = application_id

        # never redirect to an external page, this might potentially be used
        # to trick the user into thinking he's on our page after entering his
        # password and being redirected to a phising site.
        self.to = relative_url(to)

        self.yubikey_client_id = yubikey_client_id
        self.yubikey_secret_key = yubikey_secret_key

    @classmethod
    def from_app(cls, app, to='/'):
        return cls(
            session=app.session(),
            application_id=app.application_id,
            yubikey_client_id=getattr(app, 'yubikey_client_id', None),
            yubikey_secret_key=getattr(app, 'yubikey_secret_key', None),
            to=to
        )

    @classmethod
    def from_request(cls, request, to='/'):
        return cls.from_app(request.app, to)

    @classmethod
    def from_request_path(cls, request):
        return cls.from_request(request, request.transform(request.path))

    @property
    def users(self):
        return UserCollection(self.session)

    def is_valid_second_factor(self, user, second_factor_value):
        if not user.second_factor:
            return True

        if not second_factor_value:
            return False

        if user.second_factor['type'] == 'yubikey':
            return is_valid_yubikey(
                client_id=self.yubikey_client_id,
                secret_key=self.yubikey_secret_key,
                expected_yubikey_id=user.second_factor['data'],
                yubikey=second_factor_value
            )
        else:
            raise NotImplementedError

    def login(self, username, password, client='unknown', second_factor=None):
        """ Takes the given username and password and matches them against
        the users collection.

        :meth:`login_to` should be the preferred method in most cases, as it
        sets up automatic logging of login attempts and provides optional
        login/logout messages.

        :return: An identity bound to the ``application_id`` if the username
        and password match.

        """

        user = self.users.by_username_and_password(username, password)

        def fail():
            log.info("Failed login by {} ({})".format(client, username))
            return None

        if user is None:
            return fail()

        if not user.active:
            return fail()

        if not self.is_valid_second_factor(user, second_factor):
            return fail()

        log.info("Successful login by {} ({})".format(client, username))

        return self.identity_class(
            userid=user.username,
            role=user.role,
            application_id=self.application_id
        )

    def login_to(self, username, password, request, second_factor=None):
        """ Matches the username and password against the users collection,
        just like :meth:`login`. Unlike said method it returns no identity
        however, instead a redirect response is returned which will remember
        the identity.

        :return: A redirect response to ``self.to`` with the identity
        remembered as a cookie. If not successful, None is returned.

        """

        identity = self.login(username, password, request.client_addr,
                              second_factor)

        if identity is None:
            return None

        response = morepath.redirect(self.to)
        request.app.remember_identity(response, request, identity)

        return response

    def logout_to(self, request):
        """ Logs the current user out and redirects to ``self.to``.

        :return: A response redirecting to ``self.to`` with the identity
        forgotten.

        """

        response = morepath.redirect(self.to)
        request.app.forget_identity(response, request)

        return response



""" Provides commands used to manage users.

Examples:

To add a user called ``admin@example.org`` to the ``towns-govikon`` schema,
with the role 'admin'::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema add admin admin@example.org

This command will ask for a password if none was provided with ``--password``.

To delete a user::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema delete admin@example.org

To check if a user exists::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema exists admin@example.org

To change a users password::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema change-password admin@example.org

This command will also ask for a password if none was provided with
``--password``.

"""

import click

from getpass import getpass
from onegov.user import User, UserCollection
from onegov.core.cli import command_group, abort
from onegov.core.crypto import random_password

cli = command_group()


@cli.command(context_settings={'singular': True})
@click.argument('role')
@click.argument('username')
@click.option('--password', default=None,
              help="Password to give the user")
@click.option('--yubikey', default=None,
              help="The yubikey code to use for 2fa")
@click.option('--no-prompt', default=False,
              help="If no questions should be asked", is_flag=True)
def add(role, username, password, yubikey, no_prompt):
    """ Adds a user with the given name to the database. """

    def add_user(request, app):
        users = UserCollection(app.session())

        print("Adding {} to {}".format(username, app.application_id))

        if users.exists(username):
            abort("{} already exists".format(username))

        nonlocal password
        if not password:
            password = random_password(16)
            print()
            print("Using the following random password:")
            click.secho(password, fg='green')
            print()

        nonlocal yubikey
        if not yubikey and not no_prompt:
            yubikey = getpass(
                "Optionally plug in your yubi-key and press the button: "
            )

            yubikey = yubikey.strip()

        if yubikey:
            second_factor = {
                'type': 'yubikey',
                'data': yubikey[:12]
            }
        else:
            second_factor = None

        users.add(username, password, role, second_factor=second_factor)
        click.secho("{} was added".format(username), fg='green')

    return add_user


@cli.command(context_settings={'singular': True})
@click.argument('username')
def delete(username):
    """ Removes the given user from the database. """

    def delete_user(request, app):
        users = UserCollection(app.session())

        if not users.exists(username):
            abort("{} does not exist".format(username))

        users.delete(username)
        click.secho("{} was deleted".format(username), fg='green')

    return delete_user


@cli.command(context_settings={'singular': True})
@click.argument('username')
def exists(username):
    """ Returns 0 if the user exists, 1 if it doesn't. """

    def find_user(request, app):
        users = UserCollection(app.session())

        if not users.exists(username):
            abort("{} does not exist".format(username))
        else:
            click.secho("{} exists".format(username), fg='green')

    return find_user


@cli.command(context_settings={'singular': True})
@click.argument('username')
def activate(username):
    """ Activates the given user. """

    def activate_user(request, app):
        user = UserCollection(app.session()).by_username(username)

        if not user:
            abort("{} does not exist".format(username))

        user.active = True
        click.secho("{} was activated".format(username), fg='green')

    return activate_user


@cli.command(context_settings={'singular': True})
@click.argument('username')
def deactivate(username):
    """ Deactivates the given user. """

    def deactivate_user(request, app):
        user = UserCollection(app.session()).by_username(username)

        if not user:
            abort("{} does not exist".format(username))

        user.active = False
        click.secho("{} was deactivated".format(username), fg='green')

    return deactivate_user


@cli.command(context_settings={'singular': True})
@click.option('--active-only', help="Only show active users", is_flag=True)
@click.option('--inactive-only', help="Only show inactive users", is_flag=True)
def list(active_only, inactive_only):
    """ Lists all users. """

    assert not all((active_only, inactive_only))

    def list_users(request, app):

        template = '{active} {username} [{role}]'

        users = UserCollection(app.session()).query()
        users = users.with_entities(User.username, User.role, User.active)
        users = users.order_by(User.username, User.role)

        for username, role, active in users.all():
            if active_only and not active:
                continue

            if inactive_only and active:
                continue

            print(template.format(
                active=active and '✔︎' or '✘',
                username=username,
                role=role
            ))

    return list_users


@cli.command(name='change-password', context_settings={'singular': True})
@click.argument('username')
@click.option('--password', help="Password to use", default=None)
def change_password(username, password):
    """ Changes the password of the given username. """

    def change(request, app):
        users = UserCollection(app.session())

        if not users.exists(username):
            abort("{} does not exist".format(username))

        nonlocal password
        password = password or getpass("Enter password: ")

        user = users.by_username(username)
        user.password = password

        click.secho("{}'s password was changed".format(username), fg='green')

    return change


@cli.command(name='change-yubikey', context_settings={'singular': True})
@click.argument('username')
@click.option('--yubikey', help="Yubikey to use", default=None)
def change_yubikey(username, yubikey):
    """ Changes the yubikey of the given username. """

    def change(request, app):
        users = UserCollection(app.session())

        if not users.exists(username):
            abort("{} does not exist".format(username))

        nonlocal yubikey
        yubikey = (yubikey or getpass("Enter yubikey: ")).strip()[:12]

        user = users.by_username(username)
        user.second_factor = {
            'type': 'yubikey',
            'data': yubikey
        }

        click.secho("{}'s yubikey was changed".format(username), fg='green')

    return change



import os.path
import yaml

from click.testing import CliRunner
from onegov.user.cli import cli


def test_cli(postgres_dsn, session_manager, temporary_directory):

    cfg = {
        'applications': [
            {
                'path': '/foo/*',
                'application': 'onegov.core.Framework',
                'namespace': 'foo',
                'configuration': {
                    'dsn': postgres_dsn
                }
            }
        ]
    }

    session_manager.ensure_schema_exists('foo-bar')

    cfg_path = os.path.join(temporary_directory, 'onegov.yml')

    with open(cfg_path, 'w') as f:
        f.write(yaml.dump(cfg))

    runner = CliRunner()
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'add', 'admin', 'admin@example.org',
        '--password', 'hunter2',
        '--no-prompt',
    ])

    assert result.exit_code == 0
    assert 'Adding admin@example.org to foo/bar' in result.output
    assert 'admin@example.org was added' in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'add', 'admin', 'admin@example.org',
        '--password', 'hunter2',
        '--no-prompt',
    ])

    assert result.exit_code == 1
    assert 'admin@example.org already exists' in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'exists', 'admin@example.org'
    ])

    assert result.exit_code == 0
    assert 'admin@example.org exists' in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'change-password', 'admin@example.org',
        '--password', 'hunter2'
    ])

    assert result.exit_code == 0
    assert "admin@example.org's password was changed" in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'delete', 'admin@example.org'
    ])

    assert result.exit_code == 0
    assert 'admin@example.org was deleted' in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'delete', 'admin@example.org'
    ])

    assert result.exit_code == 1
    assert 'admin@example.org does not exist' in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'exists', 'admin@example.org'
    ])

    assert result.exit_code == 1
    assert 'admin@example.org does not exist' in result.output

    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'change-password', 'admin@example.org',
        '--password', 'hunter2'
    ])

    assert result.exit_code == 1
    assert 'admin@example.org does not exist' in result.output



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task
from onegov.core.orm.types import JSON
from onegov.user import User
from sqlalchemy import Boolean, Column


@upgrade_task('Add second_factor column')
def add_second_factor_column(context):
    context.operations.add_column(
        'users', Column('second_factor', JSON, nullable=True)
    )


@upgrade_task('Add active column')
def add_active_column(context):
    context.operations.add_column(
        'users', Column('active', Boolean, nullable=True, default=True)
    )

    for user in context.session.query(User).all():
        user.active = True

    context.session.flush()
    context.operations.alter_column('users', 'active', nullable=False)
